/*
 * s4diag.c	SBus BBKS-4 diagnose monitor
 *
 * Copyright (C) CWA mbH, 1992.
 *
 *
 * Author:	Michael Raus
 *
 * Date:	19.02.1992	Created
 *
 *
 * Updates:
 *	Vers	date	author	description
 *	1.10	190292	mr	created
 *	1.11	010492	mr	getting version number added
 */


#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <stdio.h>
#include <values.h>
#include <ctype.h>

#include "../link_c/link.h"

#ifdef SunOS5   /* Solaris */
#       include <unistd.h>
#       define GET_PAGE_SIZE    sysconf(_SC_PAGESIZE)
#       define USLEEP(t)        poll(0, NULL, (t/1000))
#else           /* SunOS 4.1.x */
#       define GET_PAGE_SIZE    getpagesize()
#       define USLEEP(t)        usleep(t)
#endif

static char *SccsId = "@(#)s4diag.c	1.4	95/03/24";	/* Sccs Id */


/**************************** Constant Values ******************************/

#define	MAX_FUNCTION_NO	9

#define	MAX_SBUS_SLOTS	4		/* Maximum number of SBus slots	*/

#define MENU		0
#define	SLOTS		1
#define	BOARD_RESET	2
#define	QUIT		9
#define	SBUS_ADR	0xF8000000
#define	SLOT_OFF	0x2000000
#define	RESET_OFF	0x2000

#define	MAX_BBKS4_BOARD_LINKS	4	/* Maximum number of links on a
					 * BBK-S4 board
					 */

/**************************** Type Definitions *****************************/

/* SBus descriptor */

struct sbus_descr2 {
	int 	slots;		/* number of SBus slots */
	int   	mem;		/* used address space */
	long	slot_base[MAX_SBUS_SLOTS];	/* physical SBus addresses */
};


/**************************** Global Variables *****************************/

struct sbus_descr2 sbus =  { 4, 0x8000, {0xF8000000, 0xFA000000,
					0xFC000000, 0xFE000000} };

caddr_t	page_address;		/* virtual address of page */

int	open_links[4];		/* store opened linkids */
caddr_t	link_base;
int	slot_number;
int 	no_links;		/* number of links */

/**************************** Function Prototypes **************************/

void	menu ();
void	sbslot ();

int	get_function_no ();
int	OpenSBusDevice ();
int	GetDeviceName ();
int	CloseSBusDevice ();
int	get_board_no ();
int	map_page ();
int	init_timetth ();
int	board_reset ();
int	check_links ();
char	*GetFirmwareCode ();

/***************************** External Functions **************************/

extern  int 	InfoLink  ();	/* (struct ext_link_info **LinkPtr) */
extern  char 	*lerror  ();	/* (error_code) */
extern	int	TestFirmwareCode ();	/* (LinkId) */

/***************************************************************************/

/**************************** BBK-S4 Diagnose Functions ********************/

/*
 *	menu ()
 */

void	menu ()

{	printf ("\nBBK-S4 interactive Functions:");
	printf ("\n   l    Function list");
	printf ("\n   s    SBus slots");
	printf ("\n   r    Board reset");
	printf ("\n   q    Quit\n");

	return;
}


/*
 *	get_function_no ()
 */

int	get_function_no ()

{	int	function_no;
	char	str[30];
	char	ch;

	for (;;) {
		printf ("\ns4diag ( l = function list): ");
		gets (str);
		sscanf (str, "%c", &ch);

		switch (ch) {
			case 'r':
			case 'R':
				/* reset board */
				function_no = BOARD_RESET;
				break;
			case 's':
			case 'S':
				/* list slots */
				function_no = SLOTS;
				break;
			case 'l':
			case 'L':
				/* display help */
				function_no = MENU;
				break;
			case 'q':
			case 'Q':
			case 'x':
			case 'X':
			case 'e':
			case 'E':
				/* quit */
				function_no = QUIT;
				break;
			default:
				continue;
		} /* switch ch */
		return (function_no);
	} /* for-ever */
}


/*
 *	sbslot ()
 */

void	sbslot (slotno)
int	slotno;

{	int	fd;		/* file descriptor for SBus slot */
	int	ret;		/* return value */

	if ((slotno < 0) || (slotno >= MAX_SBUS_SLOTS)) {
	     printf ("+++ Invalid slot number! Valid: 0...%d\n",
		MAX_SBUS_SLOTS-1);
	} else { 
		/* open SBus device */
		if ((fd = OpenSBusDevice (slotno)) == -1) {
		   printf ("+++ Cannot open SBus slot %d, because: ",
				slotno);
		   fflush (stdout);
		   perror("");
		} else {
		   /* read SBus device name from device EPROM and write
		    * the selected information to stdout
		    */
		   if ((ret = GetDeviceName (fd,slotno)) == -1)
			printf ("NO VALID SBUS DEVICE\n");

		   /* close opened SBus device */
		   if ((ret = CloseSBusDevice (fd)) == -1)
			printf ("+++ SBus slot %d: Device cannot be closed\n",
								      slotno);
		}
	}
}


/*	
 *	OpenSBusDevice ()
 *
 *	Opens the SBus device specified by slotnum
 *
 */

int	OpenSBusDevice (slotno)
int	slotno;

{	char	open_arg[15];	/* string containing arg to open() */
	int	fd;		/* file descriptor for SBus slot */

	/* build argument string to open */
	(void)	sprintf (open_arg, "/dev/sbus%d", slotno);

	/* open then SBus device */
	if ((fd = open (open_arg, O_RDONLY|O_NDELAY)) < 0) {
		return (-1);
	} else {
		return (fd);
	}
}



/*
 *	GetDeviceName ()
 *
 *	Read the name of the specified (slotno) SBus device
 *	from the Open Boot PROM 
 */

int	GetDeviceName (fd, slotno)
int	fd;
int	slotno;

{	int	page_size;	/* system page size in bytes */
	byte	*prom_buffer;	/* ptr to open boot PROM buffer */
	char	dev_n[50];	/* SBus device name */
	char	*dev_name;	/* ptr to SBus device name */ 	
	char	firmw[80];	/* BBK-S4 Firmware Code */
	char	*firmware;	/* ptr to BBK-S4 Firmware Code */
	char	comp_str[50];	/* compare string for Firmware Code */
	int	i, j, k, l;
	int	ret_val;
	FILE	*dummy_stream;
	u_char	dummy;

	dev_name = &dev_n[0];
	firmware = &firmw[0];
	page_size = GET_PAGE_SIZE;
	sprintf (comp_str, "%s", "CWA: BBKS4");
	link_base =  mmap (0, page_size, PROT_READ, MAP_SHARED,
				     fd, (off_t)0 );
	if (link_base == (caddr_t)-1) {
		/* mapping failed */
		fprintf (stderr, "mmap failed");
		return (-1);
	}

	dummy_stream = fdopen (fd, "r");
	ret_val = fread(&(dummy), 1, 1, dummy_stream);
 
	prom_buffer = (byte *)(link_base);

	/* read SBus board name from board EPROM */
	printf ("SBus slot %d -  ", slotno);
	
	/* check for valid FCode EPROM */
	if (dummy != 0xFD) {
		return (-1);
	}

	prom_buffer++;
	for (i = 0; i < 256; i++) {
	    if (prom_buffer[i] == 0x12) {
		j = i + 2 + (int)(prom_buffer[i+1]);
		if (((prom_buffer[j]==0x01) && (prom_buffer[j+1]==0x14)) 
		   && ((prom_buffer[j+2]==0x12) && (prom_buffer[j+3]==0x04))) { 
			j = i + 2;
			l = 0;
			for (k = 0; k < (int)prom_buffer[i+1]; k++) {
				if ((l==0) && (prom_buffer[j+k] == ' '))
					continue;
				else {
					dev_name[l] = prom_buffer[j+k];
					if (prom_buffer[j+k] == ',') {
						dev_name[l] = ':';
						dev_name[++l] = ' ';
					}
					l++;
				}
			}
			dev_name[l] = '\0';
			printf ("'%s'", dev_name);
			fflush (stdout);

			if (strcmp (dev_name, comp_str) == 0) {
				/* BBK-S4 identified */
				if ((firmware = GetFirmwareCode (slotno)) == NULL)  {
				   printf ("\nError reading Firmware Code\n\n");
				   exit (1);
				}
				printf ("  %s", firmware);
			} /* if (strcmp (dev_name, ... */
			printf ("\n");
			return (1);
		} /* if (((prom_buffer[j]==0x01) ... */
		if ((prom_buffer[j]==0x01) && (prom_buffer[j+1]==0x18)) {
			j = i + 2;
			l = 0;
			for (k = 0; k < (int)prom_buffer[i+1]; k++) {
				if ((l==0) && (prom_buffer[j+k] == ' '))
					continue;
				else {
					dev_name[l] = prom_buffer[j+k];
					if (prom_buffer[j+k] == ',') {
						dev_name[l] = ':';
						dev_name[++l] = ' ';
					}
					l++;
				}
			} /* for (k = 0; k < prom_buffer ... */
			dev_name[l] = '\0';
			printf ("'%s'", dev_name);

			if (strcmp (dev_name, comp_str) == 0) {
			   /* BBK-S4 identified */
			   if ((firmware = GetFirmwareCode (slotno)) == NULL)  {
				printf ("\nError reading Firmware Code\n\n");
				exit (1);
			   }
			   printf ("  %s", firmware);
			} /* if (strcmp (dev_name, ... */
			printf ("\n");
			return (1);
		} /* if ((prom_buffer[j]==0 ... */	
	   } /* if (prom_buffer[i] == 0x12) */
	} /* for (i = 0; i < 256; i++) ... */
	return (-1);
}


/*
 *	GetFirmwareCode ()
 */

char	*GetFirmwareCode (slotno)
int	slotno;

{static	char	firmware[80];	/* BBK-S4 Firmware Version as returned string */
	char	link_name[16];	/* OpenLink() argument */
	int	ret_val;	/* return value */
	int	link_id;	/* link identifier */
	int	loop;		/* loop counter */
    	byte	*array_on_int; 	/* byte pointer to return value */   
	struct 	ext_link_info *x_link = NULL;

    	array_on_int = (byte*) &ret_val;

	/* get number of available links */
	if ((no_links = InfoLink (&x_link)) < 0) {
		/* failed to get information about links */
		fprintf (stderr,
		    "\ns4diag: Failed to get info about links, because %s\n",
		    lerror (no_links));
		exit (1);
	}

	/* search for one free link on the actual BBK-S4 board */
	for (loop = 0; loop < no_links; loop++)  {
	    /* belongs link 'loop' to slot 'slotno' ? */
	    if (x_link[loop].slot_no == slotno)  {
		/* yes, try to get version-nr from him */
		sprintf (link_name, "%d", loop); 
		if ((ret_val = OpenLink(link_name)) == loop)  {
			/* desired link open, check him */
			if ((ret_val = TestFirmwareCode (loop)) < 0) {
			     /* failed to get firmware Version */
		    	     sprintf (firmware,
		     	     "Cannot read BBK-S4 Firmware Version, because %s",
			        lerror (ret_val) );
			} else {
				/* Firmware Version successfully read */	
				if (ret_val != VERS_NO) {
				    sprintf (firmware,
					"NOT supported Version: %d.%d",
					array_on_int[2], array_on_int[3]);
				} else {
				    /* report correct firmware Version */
				    sprintf (firmware,
					"Version: %d.%d",
					array_on_int[2], array_on_int[3]);
				} /* endif ret_val != VERS_NO */
			} /* if ((ret_val = TestFirmwareCode .. */
			/* have build answer in 'firmware' return */
			CloseLink (loop);
			return (firmware);
		} /* endif OpenLink() == loop */
	    } /* loop.slo_no == slotno */
	} /* for (loop = all-links */

	/* investigation for firmware version number failed */
	sprintf (firmware, "No link available for checking version");
	return (firmware);
}


/*
 *	CloseSBusDevice ()
 */

int	CloseSBusDevice (fd)
int	fd;

{	int	page_size;	/* system page size in bytes */
	int	ret;		/* return value */

	page_size = GET_PAGE_SIZE;

	if ( munmap ( (caddr_t) link_base, page_size) ) {
		/* failed to unmap page */
		fprintf (stderr, "Failed to unmap sbus-device\n");
		return (-1);
	}
	if ( (ret = close (fd)) != 0)
		return (-1);
	return (1);
}


/*
 *	get_board_no ()
 *	returns the read board number (slightly checked)
 */

int	get_board_no ()

{	int	board_no;

	for (;;) {
		printf ("\nReset Board : ");
		scanf ("%d", &board_no);
		for (;;) {
		    if (getchar() == '\n')
			break;
		}
		if ((board_no >= 0) && (board_no < MAX_SBUS_SLOTS))
			break;
	}
	return (board_no);
}


/*
 *	check_links ()
 *
 *	Returns the slot number (0..3) or -1 on error
 */

int	check_links (board_no, no_warning)
int	board_no;
int	no_warning;

{	struct 	ext_link_info *x_link = NULL;
	int	i;
	int	DANGER = FALSE;
	int	link_no;
	int	slot_no = -1;
	char	ch[30];
	char	link_name[16];

	if ((no_links = InfoLink (&x_link)) < 0) {
		/* failed to get info about links */
		fprintf (stderr,
		   "\nlinks: Failed to get info about links, because %s\n",
		   lerror (no_links));
		exit (1);
	}

	for (i = 0; i < 4; i++) {
		open_links[i] = -1;

		sprintf (link_name, "bbk-s4 B%d L%d", board_no, i); 
		link_no = OpenLink(link_name);

		/* does this board exist at least ? */
		if ( link_no == ENONEXIST ) {
			printf ("I can't locate BBK-S4 board %d\n", board_no);
			return (-1);
		}

		/* store the linkid */
		if ( link_no >= 0 ) {
			/* check for the same slot_no */
			if ( slot_no < 0 ) {
				slot_no = x_link[link_no].slot_no;
			} else {
			   /* check for identical slot_no's */
			   if ( slot_no != x_link[link_no].slot_no ) {
				printf ("Intern: slots link %d differ %d(%d)\n",
					link_no, x_link[link_no].slot_no, slot_no);
			   }
			   open_links[i] = link_no;
			}
		} else { /* open link ok */
			/* failed to open link 'link_name' */
			if (link_no == EINUSE) {
				/* somebody else is using this link */
				printf ("Link %d in use\n", i);
				DANGER = TRUE;
			} else if (link_no == ERNOVERSION) {
				/* failed to get firmware version */
				printf (
				 "Cannot get firmware version of link \'%s\'\n",
				 link_name);
				DANGER = TRUE;
			} else {
			    printf ("Failed to open link \'%s\', because %s\n",
				   link_name, lerror (link_no));
			}
		} /* else open link ok */
	} /* for all links on this board */

	if ( (!no_warning) && DANGER) {
		printf ("\nSome Links are in use! Reset anyway (Y/N) ? ");
		gets (ch);
		if (toupper(ch[0]) != 'Y')  {
			/* board reset aborted */
			return (-1);
		}
	}
	return (slot_no);
}


/*	
 *	map_page ()
 */

int	map_page (sbslot)
int	sbslot;

{
	int	address_offset;		/* offset from SBus base address */
	int	page_size;		/* system page size in bytes */
	off_t	page_start;		/* page-aligned address */
					/* containing address_offset */
	int	page_offset;		/* offset from page_start */
					/* to find address_offset */
	int	fd;			/* file descriptor for SBus slot */
	char	open_argument[15];	/* string containing arg to open() */
	char	*open_arg;		/* pointer to string above */


	open_arg = &open_argument[0];

	/* build argument string to open */
	(void)sprintf(open_arg, "/dev/sbus%d", sbslot);

	/* open the SBus device */
	if ((fd = open(open_arg, O_RDWR)) < 0) {
		return (-1);
	}

	/*
	 * Determine a page-aligned start address as required
	 * by mmap(). Compute the offset into that page to
	 * be mapped.
	 */

	address_offset = RESET_OFF;
	page_size = GET_PAGE_SIZE;
	page_start = (address_offset/page_size) * page_size;
	page_offset = (address_offset % page_size);

	/* Map device memory to user space */
	page_address = mmap((caddr_t)0, page_size,
			    PROT_READ|PROT_WRITE, MAP_SHARED,
			    fd, page_start);

	if (page_address == (caddr_t)-1) {
		return (-1);
	}
	return (OK);
}


/*
 *	board_reset ()
 */

int	board_reset (sbslot, no_confirm)
int	sbslot;
int	no_confirm;

{	ino_t	reset_adr;		/* board reset address */
	byte	*reset_val;		/* reset value */
	short 	tmout;			/* flag wether timeout active */
	int     timecnt;		/* timeout loop down counter */
	int	ret_val;		/* return value */
	char	ch[30];
	int	i;

	if ((ret_val = map_page (sbslot)) != OK) {
		printf("\nBoard reset: Error mapping memory page \n\n");
		return (-1);
	}

	/* compute SBus board reset address */
	reset_adr = (ino_t)(SBUS_ADR + sbslot * SLOT_OFF + RESET_OFF);
	if ( !no_confirm ) {
		printf ("\nReset Board (Y/N) ? ");
		gets (ch);
		if (toupper(ch[0]) != 'Y')
			/* board reset aborted */
			return (-1);

		printf ("Sure (Y/N) ? ");
		gets (ch);
		if (toupper(ch[0]) != 'Y')
			/* board reset aborted */
			return (-1);
	} /* if no_confirm */

	/* reset sequence */
	reset_val = (byte *)(page_address);
#ifdef	DEBUG
	printf ("\nreset address = 0x%x", reset_val);
#endif	DEBUG
	*reset_val = (byte)(0);
	(void)printf ("\n0 ...");
	*reset_val = (byte)(1);
	(void)printf (" 1 ...");
	*reset_val = (byte)(2);
	(void)printf (" 2 ...");
	*reset_val = (byte)(3);
	(void)printf (" 3 ...");
	*reset_val = (byte)(1);
	(void)printf (" 1 ...");	fflush (stdout);

	USLEEP (100000);	/* sleep about 100ms */

	(void)printf (" 0 ...");

	*reset_val = (byte)(0);

	for (i = 0; i < 4; i++)
		CloseLink(open_links[i]);

	return(OK);
}



/*
 *	usage ()
 */

static	void	usage ()

{	printf ("\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
		"Transputer board diagnostics tool",
		"usage: s4diag [-h] [-i] [-r <board-no>] [-f]",
		"	   : without any argument display SBus slots",
		"       -h : help",
		"       -i : interactive mode",
		"       -r : reset board <board-no>",
		"       -f : force board reset (not interactive)");
	exit (0);
}



/************************ BBK-S4 Diagnose Programm ************************/

int	main (argc,argv)
int	argc;
char	*argv[];

{   	int	interactive;		/* interactive flag */
   	int	force_reset;		/* force_reset flag */
	int	function;		/* selected function */
    	int	i, j;			/* loop variables */
    	int	board_no;		/* SBus slot number */
	int	ret_val;		/* return value */
    	byte	*array_on_int; 		/* byte pointer to return value */   


	if (argc > 4) usage ();    

    	array_on_int = (byte*) &ret_val;

	/* Set up default values */
	interactive = FALSE;
	force_reset = FALSE;
	board_no    = -1;
	function    = MENU;
	
	if (argc == 1)
		function = SLOTS;

	for (i = 1; i < argc; i++) {
		if (strcmp (argv[i], "-i") == 0) {
				interactive = TRUE;
		} else 	if (strcmp (argv[i], "-f") == 0) {
				force_reset = TRUE;
		} else	if (strcmp (argv[i], "-h") == 0) { 
				usage ();
		} else	if (strcmp (argv[i], "-r") == 0) {
				function = BOARD_RESET;
				/* next arg is board-no */
				if (++i >= argc) {
				   printf ("missing BBK-S4 board no\n");
				   usage ();
				} else {
				   board_no = atoi (argv[i]);
				}
		} else
			usage();
	}

	if (interactive)  {
		menu ();
		function = get_function_no ();
	}

	while (function >= 0)  {
	 switch (function)  {
	   case MENU:		/* show functions */
		menu ();
		break;
	   case SLOTS:		/* show SBus board */
		printf ("s4diag: transputer board BBK-S4 diagnostic\n");
		printf ("        s4diag -h gives help:\n\n");
		for (i = 0; i < 4; i++)  sbslot (i);
		break;
	   case BOARD_RESET: 	/* reset SBus BBK-S4 board */
		if ( interactive && ((board_no = get_board_no ()) == -1)) {
			exit (1);
		}
		if (board_no < 0) {
			printf ("Negative board numbers are not allowed!\n");
			exit (1);
		}
		if ((slot_number = check_links (board_no, force_reset)) == -1) {
			exit (1);
		}
		if ((ret_val = board_reset (slot_number, force_reset)) != OK) {
			printf("\nThe BBK-S4 board %d was NOT reseted\n\n",
				board_no);
			exit (1);
		}
		printf("\ntransputer SBus board BBK-S4 succesfully reseted \n");
			break;
	   case QUIT:		/* quit SBDiag */
			exit (0);
	   default:		/* illegal function */
			printf ("\nIllegal function");
			break;
	  } /* switch (function) */
		if (!interactive) exit (0);
		function = get_function_no ();
	} /* while (function >= 0) */
} 
