/* tcheck.c	890731		(c) Parsytec GmbH 1989
 *
 * Copyright (C) Parsytec Gmbh, 1989
 *
 * Created: July 31th 1989
 *
 * Modified:
 *
 * Description:
 *	This program demonstrates how to run a executable program
 *	on the root-transputer.
 *	The reason here is to check wether a program can be executed on
 *	the according transputer. This is done by downloading the transputer-
 *	code and then sending a byte and expecting a byte with its value
 *	incremented by one.
 */

#include "../link.h"
#include <stdio.h>

#define	BOOT_FILE	"/usr/etc/transp/trans.cde"
#define	LINK_NAME	""
#define	BUFF_SIZE	(4 * 1024)
#define	BOOT_TIMEOUT	10
#define	WRITE_TIMEOUT	10
#define	READ_TIMEOUT	10



/* external functions */
char *lerror ();	/* decodes error message */
int   OpenLink ();	/* open link */
int   ResetLink ();	/* reset transputer */
int   WriteLink ();	/* send data to transputer */
int   ReadLink  ();	/* receive data from transputer */

/* local variables */
char fbuffer[BUFF_SIZE];	/* boot-file buffer */

main (argc, argv, env)
int argc;
char *argv[], *env[];
{	FILE *fp;			/* boot-file */
	char *boot_file = BOOT_FILE;	/* transputer code file */
	char *link_name;		/* link adaptor name */
	int lerr;			/* local error code */
	int link_id;			/* used to identify the link */
	int cnt;
	int total_send;			/* bytes send so far */
	unsigned char sdata, rdata;	/* send/receive-data for transputer */

	if ( argc > 1 ) {
		/* expect link-name as parameter */
		link_name = argv[1];
	} else {
		link_name = LINK_NAME;
	}

	lerr = OpenLink (link_name);
	if ( lerr < 0 ) {
		/* failed to open link */
		fprintf (stderr,
			"\ntcheck: Failed to open link \"%s\", because %s\n",
			link_name, lerror (lerr));
		return (1);
	}
	/* store returned link-id */
	link_id = lerr;

	/* reset transputer */
	lerr = ResetLink (link_id);
	if ( lerr < 0 ) {
		/* failed to reset link */
		fprintf (stderr,
			"\ntcheck: Failed to reset link \"%s\", because %s\n",
			link_name, lerror (lerr));
		return (1);
	}

	/* download transputer code-file */
	fp = fopen (boot_file, "r");
	if ( fp == NULL ) {
		/* failed to open boot file */
		fprintf (stderr,
			"\ntcheck: Failed to open boot-file \"%s\", because ",
			boot_file);
		perror ("");
		return (1);
	}
	/* boot file opened successfully */
	total_send = 0;
	while ((cnt = fread (fbuffer, 1, sizeof (fbuffer), fp)) > 0) {
		/* try to send fbuffer to transputer */
		lerr = WriteLink (link_id, fbuffer, cnt, BOOT_TIMEOUT);
		if ( lerr != cnt ) {
			/* failed to send complete data */
			if ( lerr >= 0 ) {
				/* timeout during send */
				fprintf (stderr,
					"\ntcheck: Only %d from %d bytes send\n",
					lerr, cnt);
				return (1);
			} else {
				/* failed to send */
				fprintf (stderr,
				"\ntcheck: Failed to send at 0x%X, because %s\n",
				total_send, lerror (lerr));
				return (1);
			}
		}
		total_send += cnt;
	} /* while */
	/* check for fread error */
	if ( feof (fp) ) {
		/* end of file */
		fclose (fp);
	} else {
		/* error during read */
		fprintf (stderr,
			"\ntcheck: Read error from boot-file \"%s\",",
			boot_file);
		perror (" because ");
		return (1);
	}

	/* send single byte to transputer */
	lerr = WriteLink (link_id, &sdata, 1, WRITE_TIMEOUT);
	if ( lerr <= 0 ) {
		/* error/timeout during send */
		fprintf (stderr,
			"\ntcheck: Failed to send byte 0x%x, because %s\n",
			sdata, (lerr == 0) ? "timeout" : lerror (lerr));
		return (1);
	}

	/* receive byte incremented by one from transputer */
	lerr = ReadLink (link_id, &rdata, 1, READ_TIMEOUT);
	if ( lerr <= 0 ) {
		/* error/timeout during receive */
		fprintf (stderr,
			"\ntcheck: Failed to receive answer, because %s\n",
			(lerr == 0) ? "timeout" : lerror (lerr));
		return (1);
	}

	/* check that send++ == received */
	if ( sdata+1 != rdata ) {
		/* answer is wrong */
		fprintf (stderr,
			"\ntcheck: Have received wrong answer 0x%x/0x%x\n",
			(unsigned) rdata, (unsigned) sdata);
		return (1);
	}

	/* everything is perfectly fine. Close link */
	CloseLink (link_id);
	return (0);
}
