/* -*-C-*- */
/*
	vmeacc.h

	macros for accessing the VMEbus by a
	transputer of a bbk-v2 or by the root
	transputer of a mtm-sun

	900807 HD for parsytec
*/

#ifndef BBK_V2
#ifndef MTM_SUN

You have to define BBK_V2 or MTM_SUN

#endif
#else
#ifdef MTM_SUN

You have to define BBK_V2 or MTM_SUN

#endif
#endif


#include <string.h>

/*
	VMEbus addresses for use within C. The access
	mode is coded within the upper three or five
	address bits. You haven't to deal with that.
*/
#define _8ba		0xE0000000
#define _16ba		0xC0000000
#define _32ba		0xA0000000


/*
	With the following masks we can get the
	bits we want. You haven't to deal with that
	either.
*/
#ifdef BBK_V2
#define _8ba_mask	0x07FFFFFF
#define _16ba_mask	0x07FFFFFF
#define _32ba_mask	0x1FFFFFFF
#endif
#ifdef MTM_SUN
#define _8ba_mask	0x07FFFFFF
#define _16ba_mask	0x0FFFFFFF
#define _32ba_mask	0x1FFFFFFF
#endif


#ifdef BBK_V2
/*
	The following macro sets the address register for
	VMEbus access. This is very important !!!!
*/
#define vme_address_register(addr)	(*(int *)0x00060000 = addr)


/*
	The following macro sets the address modifier code
	for VMEbus access.

*/
#define vme_address_modifier(code)	(*(int *)0x000A0000 = code ^ 0x09)
#endif

#ifdef MTM_SUN
/*
	For the MTM-SUN address register and address modifier
	must be set within one macro call. Be carefull: width
	will be expanded more than one time.
*/
#define vme_common_address_register(addr, width, rw_control)	\
(								\
	*(int *)0x00010000=((addr) & 0xF8000000)		\
		| (rw_control)?    0x00000100: 0x00000000	\
		| ((width) == 16)? 0x00000003:			\
		  ((width) == 24)? 0x00000002: 0x00000000	\
)

#endif

/*
	8 bit access for reading and writing.
*/
#define vme_byte_8ba(addr)						\
	((char *)(_8ba + (((addr)&_8ba_mask)<<2)))

#define vme_byte_rd_8ba(var, addr)					\
	((var) = *vme_byte_8ba(addr))

#define vme_byte_wr_8ba(var, addr)					\
	(*vme_byte_8ba(addr) = (var))

#define vme_byte_rd_ar_8ba(array, size, addr)				\
	(memcpy((array), vme_byte_8ba(addr), (size)))

#define vme_byte_wr_ar_8ba(array, size, addr)				\
	(memcpy(vme_byte_8ba(addr), (array), (size)))



/*
	16 bit access for reading and writing.
*/
#ifdef BBK_V2
#define vme_int16_16ba(addr)						\
	((short *)(_16ba + (((addr)&_16ba_mask)<<2)))
#endif

#ifdef MTM_SUN
#define vme_int16_16ba(addr)						\
	((short *)(_16ba + (((addr)&_16ba_mask)<<1)))
#endif

#define vme_int16_rd_16ba(var, addr)					\
	((var) = *vme_int16_16ba(addr))

#define vme_int16_wr_16ba(var, addr)					\
	(*vme_int16_16ba(addr) = (var))

#define vme_int16_rd_ar_16ba(array, size, addr)				\
	(memcpy((array), vme_int16_16ba(addr), (size)))

#define vme_int16_wr_ar_16ba(array, size, addr)				\
	(memcpy(vme_int16_16ba(addr), (array), (size)))

/*
	32 bit access for reading and writing.
*/
#define vme_int32_32ba(addr)						\
	((int *)(_32ba+((addr)&_32ba_mask)))

#define vme_int32_rd_32ba(var, addr)					\
	((var) = *vme_int32_32ba(addr))

#define vme_int32_wr_32ba(var, addr)					\
	(*vme_int32_32ba(addr) = (var))

#define vme_int32_rd_ar_32ba(array, size, addr)				\
	(memcpy((array), vme_int32_32ba(addr), (size)))

#define vme_int32_wr_ar_32ba(array, size, addr)				\
	(memcpy(vme_int32_32ba(addr), (array), (size)))


/*
	handling 8 and 16 bit values within 32 bit access mode
*/
#define vme_byte_32ba(addr)						\
	((char *)(_32ba+((addr)&_32ba_mask)))

#define vme_byte_rd_32ba(var, addr)					\
	((var) = *vme_byte_32ba(addr))

#define vme_byte_wr_32ba(var, addr)					\
	(*vme_byte_32ba(addr) = (var))

#define vme_byte_rd_ar_32ba(array, size, addr)				\
	(memcpy((array), vme_byte_32ba(addr), (size)))

#define vme_byte_wr_ar_32ba(array, size, addr)				\
	(memcpy(vme_byte_32ba(addr), (array), (size)))


#define vme_int16_32ba(addr)						\
	((short *)(_32ba+((addr)&_32ba_mask)))

#define vme_int16_rd_32ba(var, addr)					\
	((var) = *vme_int16_32ba(addr) = (var))

#define vme_int16_wr_32ba(var, addr)					\
	(*vme_int16_32ba(addr) = (var))

#define vme_int16_rd_ar_32ba(array, size, addr)				\
	(memcpy((array), vme_int16_32ba(addr), (size)))

#define vme_int16_wr_ar_32ba(array, size, addr)				\
	(memcpy(vme_int16_32ba(addr), (array), (size)))
